let video_inverse = false  (* ou true *)
and version_caml = 0.74    (* ou 0.6 *)
and largeur_texte = 140.;; (* largeur de la page en mm *)

#open"graphics";;

type epaisseur = Fin | Moyen | Gras
and mode = Ecran | Fichier;;

let fichier_courant = ref std_out
and mode_courant = ref Ecran

and sizex = ref 0
and sizey = ref 0

and marge = 0.

and longueur_pointe = ref 0.
and largeur_pointe = ref 0.
and epaisseur = ref Fin

and min_x = ref (-.10.)
and max_x = ref 10.
and min_y = ref (-.7.5)
and max_y = ref 7.5
and delta_x = ref 20.
and delta_y = ref 15.
and largeur = ref 200.   (* en mm *)
and hauteur = ref 150.   (* en mm *)
and min_xpaper = ref 0.
and min_ypaper = ref 0.
and max_xpaper = ref 0.
and max_ypaper = ref 0.;;

let int_of_float x =
   if x >=. 0. then int_of_float x
   else let m = int_of_float(-.x) + 1 in
        int_of_float(float_of_int m +. x) - m;;

let upaper_of_u u = !largeur *. u /. !delta_x
and vpaper_of_v v = !hauteur *. v /. !delta_y
and u_of_upaper u = !delta_x *. u /. !largeur
and v_of_vpaper v = !delta_y *. v /. !hauteur
and xpaper_of_x x = !largeur *. (x -. !min_x) /. !delta_x +. marge
and ypaper_of_y y = !hauteur *. (y -. !min_y) /. !delta_y +. marge
and uscreen_of_u u =
   int_of_float( float_of_int( !sizex ) *. u /. !delta_x )
and vscreen_of_v v =
   int_of_float( float_of_int( !sizey ) *. v /. !delta_y )
and xscreen_of_x x =
   int_of_float( float_of_int( !sizex ) *. (x -. !min_x) /. !delta_x )
and yscreen_of_y y =
   int_of_float( float_of_int( !sizey ) *. (y -. !min_y) /. !delta_y );;

let output_string s = output_string !fichier_courant s;;
let output_float x =
   output_string
             (string_of_float(float_of_int(int_of_float(100. *. x)) /. 100.));;
let output_float_accolades x =
   output_string "{";
   output_float x;
   output_string "}";;
let output_floats_accolades x y =
   output_float_accolades x;
   output_float_accolades y
and output_floats x y =
   output_string "(";
   output_float x;
   output_string ",";
   output_float y;
   output_string ")"   ;;

let draw_dot x y =
   output_string"\\drawdot";
   output_floats_accolades (xpaper_of_x x) (ypaper_of_y y);
   output_string"\n";
   plot (xscreen_of_x x) (yscreen_of_y y);;

let draw_thickdot x y =
   output_string"\\drawthickdot";
   output_floats_accolades (xpaper_of_x x) (ypaper_of_y y);
   output_string"\n";
   fill_circle (xscreen_of_x x) (yscreen_of_y y) (size_x() / 200);;

let draw_hollowdot x y =
   output_string"\\drawhollowdot";
   output_floats_accolades (xpaper_of_x x) (ypaper_of_y y);
   output_string"\n";
   draw_circle (xscreen_of_x x) (yscreen_of_y y) (size_x() / 200);;

let drawpath_paper mode x y x' y' =
   output_string
      (match mode with
         0 -> "\\drawpath" |
         1 -> "\\drawdotline" |
         _ -> "\\drawdashline");
   output_floats_accolades x y;
   output_floats_accolades x' y';
   output_string"\n";;

let drawpath mode x y x' y' =
   drawpath_paper mode
      (xpaper_of_x x) (ypaper_of_y y)
      (xpaper_of_x x') (ypaper_of_y y');
   moveto (xscreen_of_x x) (yscreen_of_y y);
   lineto (xscreen_of_x x') (yscreen_of_y y');;

let draw_line = drawpath 0
and draw_dotline = drawpath 1
and draw_dashline = drawpath 2;;

let couplepaper_correct x y =
   x >=. !min_xpaper & x <=. !max_xpaper & y >=. !min_ypaper & y <=. !max_ypaper;;

let path_paper =
   let rec f compteur = function
      x :: y :: l as l' ->
         if couplepaper_correct x y
            then
               match compteur with
                  0 ->  output_floats x y;
                        output_string"\n";
                        f 10 l' |
                  10 -> output_string"\\path";
                        output_floats x y;
                        f 2 l |
                  _ ->  output_floats x y;
                        f ( (compteur + 1) mod 10 ) l
            else
               (
               if compteur < 10 then output_string"\n";
               f 10 l
               ) |
      _ -> if compteur < 10 then output_string"\n" in
      f 10;;

let draw_path l =
   let rec f premier = function
      x :: y :: l -> if premier then moveto (xscreen_of_x x) (yscreen_of_y y)
                                else lineto (xscreen_of_x x) (yscreen_of_y y);
                     (xpaper_of_x x) :: (ypaper_of_y y) :: (f false l) |
      _ -> [] in
   path_paper(f true l);;

let thinlines() =
   set_line_width 1;
   epaisseur := Fin;
   longueur_pointe := 1.;
   largeur_pointe := 0.7;
   output_string"\\thinlines\n"
and thicklines() =
   set_line_width 2;
   epaisseur := Moyen;
   longueur_pointe := 1.1;
   largeur_pointe := 0.8;
   output_string"\\thicklines\n"
and Thicklines() =
   set_line_width 3;
   epaisseur := Gras;
   longueur_pointe := 1.2;
   largeur_pointe := 0.9;
   output_string"\\Thicklines\n";;

let drawvector_paper corps x y u v =
   let norme = sqrt(u *. u +. v *. v) in
   let (x0,y0) = if corps then (x +. u,y +. v) else (x,y)
   and u = u /. norme
   and v = v /. norme in
   let u' = v *. !largeur_pointe *. 0.5
   and v' = -. u *. !largeur_pointe *. 0.5
   and x' = x0 -. !longueur_pointe *. u
   and y' = y0 -. !longueur_pointe *. v in
   let x1 = x' +. u'
   and y1 = y' +. v'
   and x2 = x' -. u'
   and y2 = y' -. v' in
   let sauve_epaisseur = !epaisseur in
   Thicklines();
   path_paper[x0;y0;x1;y1;x2;y2;x0;y0];
   (match sauve_epaisseur with
      Fin -> thinlines() |
      Moyen -> thicklines() |
      _ -> ());
   if corps then drawpath_paper 0 x y x' y';;

let drawvector_screen corps x y u v =
   let norme = sqrt(float_of_int(u * u + v * v)) in
   let (x0,y0) = if corps then (x + u,y + v) else (x,y)
   and u = int_of_float(float_of_int(10 * u) /. norme)
   and v = int_of_float(float_of_int(10 * v) /. norme) in
   let u' = v / 3
   and v' = - u / 3
   and x' = x0 - u
   and y' = y0 - v in
   let x1 = x' + u'
   and y1 = y' + v'
   and x2 = x' - u'
   and y2 = y' - v' in
   moveto x0 y0;
   lineto x1 y1;
   lineto x2 y2;
   lineto x0 y0;
   if corps then lineto x y;;

let drawvector corps x y u v =
   drawvector_paper
     corps (xpaper_of_x x) (ypaper_of_y y) (upaper_of_u u) (vpaper_of_v v);
   drawvector_screen
     corps (xscreen_of_x x) (yscreen_of_y y) (uscreen_of_u u) (vscreen_of_v v);;

let draw_vector = drawvector true
and draw_arrow = drawvector false;;

let draw_curve x y min max nombrepoints =
   let pas = (max -. min) /. (float_of_int( nombrepoints - 1 )) in
   let rec f t = if t >. max +. pas /. 2.
      then []
      else x t :: y t :: f (t +. pas) in
   draw_path(f min);;

let draw_text_screen mode x y texte =
   let (l,h) = text_size texte in
   moveto (match mode with
                      0 -> x - l / 2 |
                      1 -> x |
                      _ -> x - l) (y - h / 2);
   draw_string texte;;

let draw_text mode x y texte =
   output_string("\\draw" ^ (match mode with
                                 0 -> "centered" |
                                 1 -> "left" |
                                 _ -> "right") ^ "text");
   output_floats_accolades (xpaper_of_x x) (ypaper_of_y y);
   output_string ("{" ^ texte ^ "}\n");
   draw_text_screen mode (xscreen_of_x x) (yscreen_of_y y) texte;;

let draw_centeredtext = draw_text 0
and draw_lefttext = draw_text 1
and draw_righttext = draw_text 2;;

let actionshade_screen shade action =
    let s = int_of_float
               (255. *. (if video_inverse then shade else 1. -. shade)) in
    set_color (rgb s s s);
    action();
    set_color (if video_inverse then white else black);;

let draw_shadedcircle x_centre y_centre diametre texte shade =
   (* diametre horizontal, 0. <=. shade <=. 1. *)
   output_string ("\\draw" ^ (if shade <. 0. then "" else "shaded") ^ "circle");
   output_floats_accolades (xpaper_of_x x_centre) (ypaper_of_y y_centre);
   output_float_accolades (upaper_of_u diametre);
   output_string ("{" ^ texte ^ "}");
   if shade >=. 0. then output_float_accolades shade;
   output_string "\n";
   let rayon = (uscreen_of_u diametre) / 2
   and x = xscreen_of_x x_centre
   and y = yscreen_of_y y_centre in
   if shade >=. 0.
      then actionshade_screen shade (function () -> fill_circle x y rayon);
   draw_circle x y rayon;
   draw_text_screen 0 x y texte;;

let draw_circle x_centre y_centre diametre texte =
   draw_shadedcircle x_centre y_centre diametre texte (-. 1.);;

let draw_shadedbox x_centre y_centre largeur hauteur texte shade =
   (* 0. <=. shade <=. 1. *)
   let s = shade >=. 0.
   and minx = x_centre -. largeur /. 2.
   and miny = y_centre -. hauteur /. 2. in
   output_string ("\\draw" ^ (if s then "shade" else "frame") ^ "box");
   if s then output_float_accolades (xpaper_of_x minx);
   output_float_accolades (xpaper_of_x x_centre);
   if s then output_floats_accolades
                (xpaper_of_x (x_centre +. largeur /. 2.)) (ypaper_of_y miny);
   output_float_accolades (ypaper_of_y y_centre);
   if s
      then output_float_accolades (ypaper_of_y (y_centre +. hauteur /. 2.))
      else output_floats_accolades (upaper_of_u largeur) (vpaper_of_v hauteur);
   output_string ("{" ^ texte ^ "}");
   if s then output_float_accolades shade;
   output_string "\n";
   let l = uscreen_of_u largeur
   and h = vscreen_of_v hauteur
   and x = xscreen_of_x minx
   and y = yscreen_of_y miny in
   if s then actionshade_screen shade (function () -> fill_rect x y l h);
   moveto x y;
   lineto (x + l) y;
   lineto (x + l) (y + h);
   lineto x (y + h);
   lineto x y;
   draw_text_screen 0 (x + l / 2) (y + h / 2) texte;;

let draw_box x_centre y_centre largeur hauteur texte =
   draw_shadedbox x_centre y_centre largeur hauteur texte (-.1.);;

let draw_eps fichier largeur hauteur x_min y_max =
   let x_centre = x_min +. largeur /. 2.
   and y_centre = y_max -. hauteur /. 2. in
   output_string ("\\drawpostscript");
   output_floats_accolades (xpaper_of_x x_centre) (ypaper_of_y y_centre);
   output_floats_accolades (upaper_of_u largeur) (vpaper_of_v hauteur);
   output_string ("{" ^ fichier ^ "}\n");
   actionshade_screen 0.4
      (function () ->
         fill_rect (xscreen_of_x x_min) (yscreen_of_y (y_max -. hauteur))
                   (uscreen_of_u largeur) (vscreen_of_v hauteur));;

let open_graph fichier minx maxx miny maxy larg haut =
   open_graph"";
   if fichier = ""
      then
         (
         mode_courant := Ecran;
         fichier_courant := std_out
         )
      else
         (
         mode_courant := Fichier;
         fichier_courant := open_out fichier
         );
   min_x := minx;
   max_x := maxx;
   min_y := miny;
   max_y := maxy;
   delta_x := maxx -. minx;
   delta_y := maxy -. miny;
   largeur := larg;
   hauteur := haut;
   if float_of_int(size_x()) *. haut <. float_of_int(size_y()) *. larg
      then
         (
         sizex := size_x();
         sizey := int_of_float(float_of_int(size_x()) *. haut /. larg);
         moveto 0 (!sizey - 1)
         )
      else
         (
         sizey := size_y();
         sizex := int_of_float(float_of_int(size_y()) *. larg /. haut);
         moveto (!sizex - 1) 0
         );
   lineto (!sizex - 1) (!sizey - 1);
   min_xpaper := xpaper_of_x !min_x;
   min_ypaper := ypaper_of_y !min_y;
   max_xpaper := xpaper_of_x !max_x;
   max_ypaper := ypaper_of_y !max_y;
   output_string"\\begin{picture}";
   output_floats (!largeur +. 2. *. marge) (!hauteur +. 2. *. marge);
   output_string"\n";
   thinlines();;

let close_graph() =
   output_string"\\end{picture}";
   close_out !fichier_courant;;

let pi = 3.141592654;;

let acos =
 let rec ac x = (* pas rapide mais plus prcis que la fct acos de CAML *)
   if x <. 0.
      then
         pi -. ac (-.x)
      else
         let rec sol a b ca cb =
            if abs_float(b -. a) <. 1e-8
               then a
               else let c = (a +. b) /. 2. in
                    let cc = cos c in
                    if cc <. x
                       then sol a c ca cc
                       else sol c b cc cb in
         sol 0. (pi/.2.) 1. 0. in
 ac;;
(*if version_caml = 0.6 then ac else acos;;*)

let draw_circlearc xcentre ycentre diametre t2 t1 =
   let dpi = 2. *. pi in
   let r t = int_of_float(t *. 360. /. dpi)
   and normal t = t -. float_of_int(int_of_float(t /. dpi)) *. dpi in
   let t1' = normal (-.t1)
   and t2' = normal (-.t2) in
   let t2'' = if t2' >=. t1' then t2' else t2' +. dpi
   and rayonscreen = uscreen_of_u diametre / 2 in
   let t1s,t2s =
      if version_caml = 0.6
         then t1',t2''
         else let ts1' = normal t2
              and ts2' = normal t1 in
              ts1',(if ts2' >=. ts1' then ts2' else ts2' +. dpi) in
   graphics__draw_arc (xscreen_of_x xcentre) (yscreen_of_y ycentre)
                      rayonscreen rayonscreen (r t1s) (r t2s);
   let xp = xpaper_of_x xcentre
   and yp = ypaper_of_y ycentre
   and dp = upaper_of_u diametre in
   output_string"\\drawarc";
   output_floats_accolades xp yp;
   output_float_accolades dp;
   output_floats_accolades t1' t2'';
   output_string"\n";;
(*

let draw_circlearc x y d t t' = (* version plus prcise mais plus longue *)
   let dpi = 2. *. pi in
   let normal t = t -. float_of_int(int_of_float(t /. dpi)) *. dpi in
   let t1 = normal t
   and t2 = normal t' in
   let t2 = if t2 >=. t1 then t2 else t2 +. dpi
   and r = d/.2. in
   draw_curve (function t-> x +. r*.cos t)
             (function t-> y +. r*.sin t)
             t1 t2 40;;
*)

let norme (u,v) = sqrt(u *. u +. v *. v);;

let normer (u,v) = let n = norme (u,v) in (u /. n,v /. n);;

type position = G | M | D
and vertex == float * float * float;;

let interlignes2 = 1.5 and ecart_etiquettearc = 0.5;;  (* mms *)

let draw_etiquettearc x y u v position texte =
   if position = M
      then
         draw_centeredtext x y texte
      else
         let il = u_of_upaper interlignes2
         and ee = u_of_upaper ecart_etiquettearc
         and (u,v) = normer(if position = D then (u,v) else (-.u,-.v)) in
         if abs_float v <. 0.1
            then
               draw_centeredtext
                  x (if u >. 0. then y -.ee-.il else y+.ee+.il) texte
         else
            let f,x' =
               if v >. 0.
                  then draw_lefttext,x +. ee
                  else draw_righttext,x -. ee
            and y' =
               if abs_float u <. 0.1
                  then y
                  else if u >. 0. then y -. il else y +. il in
            f x' y' texte;;

let draw_vertex x y diametre etiquette =
  draw_circle x y diametre etiquette;
  x,y,diametre;;

let draw_arc_hpos
   nb_fleches (x1, y1, diametre1) (x2, y2, diametre2) h position etiquette =
   let a,b = normer(x2 -. x1,y2 -. y1) in
   let x0 = (x1 +. x2) /. 2.
   and y0 = (y1 +. y2) /. 2.
   and u_of_UV (U,V) = a *. U -. b *. V
   and v_of_UV (U,V) = b *. U +. a *. V in
   let x_of_XY (X,Y) = x0 +. u_of_UV (X,Y)
   and y_of_XY (X,Y) = y0 +. v_of_UV (X,Y)
   and theta = if b >=. 0. then acos a else -.acos a in
   let t_of_T T = T +. theta
   and l = norme (x2 -. x1,y2 -. y1) /. 2. in
   let YC = (h *. h -. l *. l) *. 0.5 /. h in
   let R = abs_float(YC -. h)
   and R1 = diametre1 *. 0.5
   and R2 = diametre2 *. 0.5 in
   let phik r = acos(1. -. 0.5 *. (r *. r) /. (R *. R)) in
   let phi1 = phik R1
   and phi2 = phik R2
   and phi = -. asin(YC /. R) in
   let Tmin = phi +. phi2
   and Tmax = acos(-.1.) -. phi -. phi1 in
   let tmin = t_of_T Tmin
   and tmax = t_of_T Tmax in
   draw_circlearc (x_of_XY(0.,YC)) (y_of_XY(0.,YC)) (2. *. R) tmin tmax;
   let da t v =
      let c = R *. cos t
      and s = R *. sin t in
      let N = c,s +. YC
      and V = -.v *. s,v *. c in
      draw_arrow (x_of_XY N) (y_of_XY N) (u_of_UV V) (v_of_UV V) in
   if nb_fleches >= 1 then da Tmin (-.1.);
   if nb_fleches = 2 or nb_fleches = -1 then da Tmax 1.;
   draw_etiquettearc (x_of_XY(0.,h)) (y_of_XY(0.,h)) a b position etiquette;;

let draw_arc
   nb_fleches h ((x1, y1, diametre1) as s1) ((x2, y2, diametre2) as s2)
                                                     position etiquette =
   if h = 0.
      then
         let a,b = normer(x2 -. x1,y2 -. y1) in
         let x1' = x1 +. a *. diametre1 /. 2.
         and y1' = y1 +. b *. diametre1 /. 2.
         and x2' = x2 -. a *. diametre2 /. 2.
         and y2' = y2 -. b *. diametre2 /. 2. in
         draw_line x1' y1' x2' y2';
         if nb_fleches >= 1 then draw_arrow x2' y2' a b;
         if nb_fleches = 2 then draw_arrow x1' y1' (-.a) (-.b);
         draw_etiquettearc ((x1'+.x2')/.2.) ((y1'+.y2')/.2.) a b
                                                         position etiquette
      else
         if h >. 0.
            then
               draw_arc_hpos nb_fleches s1 s2 h position etiquette
            else
               draw_arc_hpos
                  (if nb_fleches = 1 then -1 else nb_fleches)
                  s2 s1 (-.h)
                  (match position with G -> D|D -> G|M -> M) etiquette;;

let draw_inout entrant longueur angle ((x,y,diametre) as s) =
   let l = diametre /. 2. +. longueur in
   let s' = x +. l *. cos angle , y +. l *. sin angle , 0. in
   let s1,s2 = if entrant then s',s else s,s' in
   draw_arc 1 0. s1 s2 M "";;

let draw_in = draw_inout true
and draw_out = draw_inout false;;

let draw_loop nombre_fleches diam angle  (x,y,diametre) position etiquette =
   let r = diametre /. 2.
   and rho = diam /. 2.
   and pi = acos(-.1.) in
   let l = if r <. rho then rho +. r /. 2. else r +. rho /. 2. in
   let phi = acos((rho *. rho +. l *. l -. r *. r) /. (2. *. rho *. l))
   and ca = cos angle
   and sa = sin angle in
   let x' = x +. l *. ca
   and y' = y +. l *. sa
   and tmin = angle +. phi -. pi
   and tmax = angle +. pi -. phi in
   draw_circlearc x' y' diam tmin tmax;
   let c = rho *. cos tmax
   and s = rho *. sin tmax in
   if nombre_fleches = 1 then draw_arrow (x' +. c) (y' +. s) (-.s) c;
   draw_etiquettearc (x'+. rho *. ca) (y'+. rho *. sa) (-.sa) ca
                                                         position etiquette;;
